// src/babel-esm.ts
import { parse } from "@babel/parser";
import * as t from "@babel/types";
import { createRequire } from "module";
var require2 = createRequire(import.meta.url);
var _traverse = require2("@babel/traverse");
var traverse = _traverse.default;

// src/identifier.ts
function fromFunction(path) {
  let parentPath = path.parentPath;
  if (parentPath.type === "VariableDeclarator") {
    let variablePath = parentPath;
    let name = variablePath.get("id");
    return name.node.type === "Identifier" ? name : null;
  }
  if (parentPath.type === "AssignmentExpression") {
    let variablePath = parentPath;
    let name = variablePath.get("left");
    return name.node.type === "Identifier" ? name : null;
  }
  if (path.node.type === "ArrowFunctionExpression") {
    return null;
  }
  return path.node.id && path.node.id.type === "Identifier" ? path.get("id") : null;
}
function isReferenced(ident) {
  let binding = ident.scope.getBinding(ident.node.name);
  if (binding?.referenced) {
    if (binding.path.type === "FunctionDeclaration") {
      return !binding.constantViolations.concat(binding.referencePaths).every((ref) => ref.findParent((parent) => parent === binding?.path));
    }
    return true;
  }
  return false;
}

// src/dead-code-elimination.ts
function dead_code_elimination_default(ast, candidates) {
  let referencesRemovedInThisPass;
  let shouldBeRemoved = (ident) => {
    if (isReferenced(ident))
      return false;
    if (!candidates)
      return true;
    return candidates.has(ident);
  };
  let sweepFunction = (path) => {
    let identifier = fromFunction(path);
    if (identifier?.node && shouldBeRemoved(identifier)) {
      ++referencesRemovedInThisPass;
      if (t.isAssignmentExpression(path.parentPath.node) || t.isVariableDeclarator(path.parentPath.node)) {
        path.parentPath.remove();
      } else {
        path.remove();
      }
    }
  };
  let sweepImport = (path) => {
    let local = path.get("local");
    if (shouldBeRemoved(local)) {
      ++referencesRemovedInThisPass;
      path.remove();
      if (path.parent.specifiers.length === 0) {
        path.parentPath.remove();
      }
    }
  };
  do {
    referencesRemovedInThisPass = 0;
    traverse(ast, {
      Program(path) {
        path.scope.crawl();
      },
      // eslint-disable-next-line no-loop-func
      VariableDeclarator(path) {
        if (path.node.id.type === "Identifier") {
          let local = path.get("id");
          if (shouldBeRemoved(local)) {
            ++referencesRemovedInThisPass;
            path.remove();
          }
        } else if (path.node.id.type === "ObjectPattern") {
          let pattern = path.get("id");
          let beforeCount = referencesRemovedInThisPass;
          let properties = pattern.get("properties");
          properties.forEach((property) => {
            let local = property.get(
              property.node.type === "ObjectProperty" ? "value" : property.node.type === "RestElement" ? "argument" : function() {
                throw new Error("invariant");
              }()
            );
            if (shouldBeRemoved(local)) {
              ++referencesRemovedInThisPass;
              property.remove();
            }
          });
          if (beforeCount !== referencesRemovedInThisPass && pattern.get("properties").length < 1) {
            path.remove();
          }
        } else if (path.node.id.type === "ArrayPattern") {
          let pattern = path.get("id");
          let beforeCount = referencesRemovedInThisPass;
          let elements = pattern.get("elements");
          elements.forEach((e) => {
            let local;
            if (e.node?.type === "Identifier") {
              local = e;
            } else if (e.node?.type === "RestElement") {
              local = e.get("argument");
            } else {
              return;
            }
            if (shouldBeRemoved(local)) {
              ++referencesRemovedInThisPass;
              e.remove();
            }
          });
          if (beforeCount !== referencesRemovedInThisPass && pattern.get("elements").length < 1) {
            path.remove();
          }
        }
      },
      FunctionDeclaration: sweepFunction,
      FunctionExpression: sweepFunction,
      ArrowFunctionExpression: sweepFunction,
      ImportSpecifier: sweepImport,
      ImportDefaultSpecifier: sweepImport,
      ImportNamespaceSpecifier: sweepImport
    });
  } while (referencesRemovedInThisPass);
}

// src/find-referenced-identifiers.ts
function find_referenced_identifiers_default(ast) {
  const referenced = /* @__PURE__ */ new Set();
  function markFunction(path) {
    const ident = fromFunction(path);
    if (ident?.node && isReferenced(ident)) {
      referenced.add(ident);
    }
  }
  function markImport(path) {
    const local = path.get("local");
    if (isReferenced(local)) {
      referenced.add(local);
    }
  }
  traverse(ast, {
    VariableDeclarator(path) {
      if (path.node.id.type === "Identifier") {
        const local = path.get("id");
        if (isReferenced(local)) {
          referenced.add(local);
        }
      } else if (path.node.id.type === "ObjectPattern") {
        const pattern = path.get("id");
        const properties = pattern.get("properties");
        properties.forEach((p) => {
          const local = p.get(
            p.node.type === "ObjectProperty" ? "value" : p.node.type === "RestElement" ? "argument" : function() {
              throw new Error("invariant");
            }()
          );
          if (isReferenced(local)) {
            referenced.add(local);
          }
        });
      } else if (path.node.id.type === "ArrayPattern") {
        const pattern = path.get("id");
        const elements = pattern.get("elements");
        elements.forEach((e) => {
          let local;
          if (e.node?.type === "Identifier") {
            local = e;
          } else if (e.node?.type === "RestElement") {
            local = e.get("argument");
          } else {
            return;
          }
          if (isReferenced(local)) {
            referenced.add(local);
          }
        });
      }
    },
    FunctionDeclaration: markFunction,
    FunctionExpression: markFunction,
    ArrowFunctionExpression: markFunction,
    ImportSpecifier: markImport,
    ImportDefaultSpecifier: markImport,
    ImportNamespaceSpecifier: markImport
  });
  return referenced;
}
export {
  dead_code_elimination_default as deadCodeElimination,
  find_referenced_identifiers_default as findReferencedIdentifiers
};
